@@ -116,7 +116,7 @@ uint64_t uv_hrtime(void) {
 void uv_close(uv_handle_t* handle, uv_close_cb close_cb) {
   assert(!uv__is_closing(handle));
 
-  handle->flags |= UV_CLOSING;
+  handle->flags |= UV_HANDLE_CLOSING;
   handle->close_cb = close_cb;
 
   switch (handle->type) {
@@ -174,8 +174,8 @@ void uv_close(uv_handle_t* handle, uv_close_cb close_cb) {
 
   case UV_SIGNAL:
     uv__signal_close((uv_signal_t*) handle);
-    /* Signal handles may not be closed immediately. The signal code will */
-    /* itself close uv__make_close_pending whenever appropriate. */
+    /* Signal handles may not be closed immediately. The signal code will
+     * itself close uv__make_close_pending whenever appropriate. */
     return;
 
   default:
@@ -191,14 +191,14 @@ int uv__socket_sockopt(uv_handle_t* handle, int optname, int* value) {
   socklen_t len;
 
   if (handle == NULL || value == NULL)
-    return -EINVAL;
+    return UV_EINVAL;
 
   if (handle->type == UV_TCP || handle->type == UV_NAMED_PIPE)
     fd = uv__stream_fd((uv_stream_t*) handle);
   else if (handle->type == UV_UDP)
     fd = ((uv_udp_t *) handle)->io_watcher.fd;
   else
-    return -ENOTSUP;
+    return UV_ENOTSUP;
 
   len = sizeof(*value);
 
@@ -208,14 +208,14 @@ int uv__socket_sockopt(uv_handle_t* handle, int optname, int* value) {
     r = setsockopt(fd, SOL_SOCKET, optname, (const void*) value, len);
 
   if (r < 0)
-    return -errno;
+    return UV__ERR(errno);
 
   return 0;
 }
 
 void uv__make_close_pending(uv_handle_t* handle) {
-  assert(handle->flags & UV_CLOSING);
-  assert(!(handle->flags & UV_CLOSED));
+  assert(handle->flags & UV_HANDLE_CLOSING);
+  assert(!(handle->flags & UV_HANDLE_CLOSED));
   handle->next_closing = handle->loop->closing_handles;
   handle->loop->closing_handles = handle;
 }
@@ -241,15 +241,17 @@ int uv__getiovmax(void) {
 
 
 static void uv__finish_close(uv_handle_t* handle) {
-  /* Note: while the handle is in the UV_CLOSING state now, it's still possible
-   * for it to be active in the sense that uv__is_active() returns true.
+  /* Note: while the handle is in the UV_HANDLE_CLOSING state now, it's still
+   * possible for it to be active in the sense that uv__is_active() returns
+   * true.
+   *
    * A good example is when the user calls uv_shutdown(), immediately followed
    * by uv_close(). The handle is considered active at this point because the
    * completion of the shutdown req is still pending.
    */
-  assert(handle->flags & UV_CLOSING);
-  assert(!(handle->flags & UV_CLOSED));
-  handle->flags |= UV_CLOSED;
+  assert(handle->flags & UV_HANDLE_CLOSING);
+  assert(!(handle->flags & UV_HANDLE_CLOSED));
+  handle->flags |= UV_HANDLE_CLOSED;
 
   switch (handle->type) {
     case UV_PREPARE:
@@ -418,12 +420,12 @@ int uv__socket(int domain, int type, int protocol) {
     return sockfd;
 
   if (errno != EINVAL)
-    return -errno;
+    return UV__ERR(errno);
 #endif
 
   sockfd = socket(domain, type, protocol);
   if (sockfd == -1)
-    return -errno;
+    return UV__ERR(errno);
 
   err = uv__nonblock(sockfd, 1);
   if (err == 0)
@@ -487,7 +489,7 @@ int uv__accept(int sockfd) {
       continue;
 
     if (errno != ENOSYS)
-      return -errno;
+      return UV__ERR(errno);
 
     no_accept4 = 1;
 skip:
@@ -497,7 +499,7 @@ skip:
     if (peerfd == -1) {
       if (errno == EINTR)
         continue;
-      return -errno;
+      return UV__ERR(errno);
     }
 
     err = uv__cloexec(peerfd, 1);
@@ -523,8 +525,8 @@ int uv__close_nocheckstdio(int fd) {
   saved_errno = errno;
   rc = close(fd);
   if (rc == -1) {
-    rc = -errno;
-    if (rc == -EINTR || rc == -EINPROGRESS)
+    rc = UV__ERR(errno);
+    if (rc == UV_EINTR || rc == UV__ERR(EINPROGRESS))
       rc = 0;    /* The close is in progress, not an error. */
     errno = saved_errno;
   }
@@ -536,7 +538,7 @@ int uv__close_nocheckstdio(int fd) {
 int uv__close(int fd) {
   assert(fd > STDERR_FILENO);  /* Catch stdio close bugs. */
 #if defined(__MVS__)
-  epoll_file_close(fd);
+  SAVE_ERRNO(epoll_file_close(fd));
 #endif
   return uv__close_nocheckstdio(fd);
 }
@@ -550,7 +552,7 @@ int uv__nonblock_ioctl(int fd, int set) {
   while (r == -1 && errno == EINTR);
 
   if (r)
-    return -errno;
+    return UV__ERR(errno);
 
   return 0;
 }
@@ -565,7 +567,7 @@ int uv__cloexec_ioctl(int fd, int set) {
   while (r == -1 && errno == EINTR);
 
   if (r)
-    return -errno;
+    return UV__ERR(errno);
 
   return 0;
 }
@@ -581,7 +583,7 @@ int uv__nonblock_fcntl(int fd, int set) {
   while (r == -1 && errno == EINTR);
 
   if (r == -1)
-    return -errno;
+    return UV__ERR(errno);
 
   /* Bail out now if already set/clear. */
   if (!!(r & O_NONBLOCK) == !!set)
@@ -597,7 +599,7 @@ int uv__nonblock_fcntl(int fd, int set) {
   while (r == -1 && errno == EINTR);
 
   if (r)
-    return -errno;
+    return UV__ERR(errno);
 
   return 0;
 }
@@ -612,7 +614,7 @@ int uv__cloexec_fcntl(int fd, int set) {
   while (r == -1 && errno == EINTR);
 
   if (r == -1)
-    return -errno;
+    return UV__ERR(errno);
 
   /* Bail out now if already set/clear. */
   if (!!(r & FD_CLOEXEC) == !!set)
@@ -628,7 +630,7 @@ int uv__cloexec_fcntl(int fd, int set) {
   while (r == -1 && errno == EINTR);
 
   if (r)
-    return -errno;
+    return UV__ERR(errno);
 
   return 0;
 }
@@ -643,7 +645,7 @@ int uv__dup(int fd) {
   fd = dup(fd);
 
   if (fd == -1)
-    return -errno;
+    return UV__ERR(errno);
 
   err = uv__cloexec(fd, 1);
   if (err) {
@@ -667,10 +669,10 @@ ssize_t uv__recvmsg(int fd, struct msghdr* msg, int flags) {
     if (rc != -1)
       return rc;
     if (errno != EINVAL)
-      return -errno;
+      return UV__ERR(errno);
     rc = recvmsg(fd, msg, flags);
     if (rc == -1)
-      return -errno;
+      return UV__ERR(errno);
     no_msg_cmsg_cloexec = 1;
   } else {
     rc = recvmsg(fd, msg, flags);
@@ -679,7 +681,7 @@ ssize_t uv__recvmsg(int fd, struct msghdr* msg, int flags) {
   rc = recvmsg(fd, msg, flags);
 #endif
   if (rc == -1)
-    return -errno;
+    return UV__ERR(errno);
   if (msg->msg_controllen == 0)
     return rc;
   for (cmsg = CMSG_FIRSTHDR(msg); cmsg != NULL; cmsg = CMSG_NXTHDR(msg, cmsg))
@@ -695,10 +697,10 @@ ssize_t uv__recvmsg(int fd, struct msghdr* msg, int flags) {
 
 int uv_cwd(char* buffer, size_t* size) {
   if (buffer == NULL || size == NULL)
-    return -EINVAL;
+    return UV_EINVAL;
 
   if (getcwd(buffer, *size) == NULL)
-    return -errno;
+    return UV__ERR(errno);
 
   *size = strlen(buffer);
   if (*size > 1 && buffer[*size - 1] == '/') {
@@ -712,7 +714,7 @@ int uv_cwd(char* buffer, size_t* size) {
 
 int uv_chdir(const char* dir) {
   if (chdir(dir))
-    return -errno;
+    return UV__ERR(errno);
 
   return 0;
 }
@@ -749,11 +751,11 @@ int uv_fileno(const uv_handle_t* handle, uv_os_fd_t* fd) {
     break;
 
   default:
-    return -EINVAL;
+    return UV_EINVAL;
   }
 
   if (uv__is_closing(handle) || fd_out == -1)
-    return -EBADF;
+    return UV_EBADF;
 
   *fd = fd_out;
   return 0;
@@ -927,11 +929,16 @@ int uv__io_active(const uv__io_t* w, unsigned int events) {
 }
 
 
+int uv__fd_exists(uv_loop_t* loop, int fd) {
+  return (unsigned) fd < loop->nwatchers && loop->watchers[fd] != NULL;
+}
+
+
 int uv_getrusage(uv_rusage_t* rusage) {
   struct rusage usage;
 
   if (getrusage(RUSAGE_SELF, &usage))
-    return -errno;
+    return UV__ERR(errno);
 
   rusage->ru_utime.tv_sec = usage.ru_utime.tv_sec;
   rusage->ru_utime.tv_usec = usage.ru_utime.tv_usec;
@@ -973,7 +980,7 @@ int uv__open_cloexec(const char* path, int flags) {
       return fd;
 
     if (errno != EINVAL)
-      return -errno;
+      return UV__ERR(errno);
 
     /* O_CLOEXEC not supported. */
     no_cloexec = 1;
@@ -982,7 +989,7 @@ int uv__open_cloexec(const char* path, int flags) {
 
   fd = open(path, flags);
   if (fd == -1)
-    return -errno;
+    return UV__ERR(errno);
 
   err = uv__cloexec(fd, 1);
   if (err) {
@@ -999,14 +1006,14 @@ int uv__dup2_cloexec(int oldfd, int newfd) {
 #if (defined(__FreeBSD__) && __FreeBSD__ >= 10) || defined(__NetBSD__)
   r = dup3(oldfd, newfd, O_CLOEXEC);
   if (r == -1)
-    return -errno;
+    return UV__ERR(errno);
   return r;
 #elif defined(__FreeBSD__) && defined(F_DUP2FD_CLOEXEC)
   r = fcntl(oldfd, F_DUP2FD_CLOEXEC, newfd);
   if (r != -1)
     return r;
   if (errno != EINVAL)
-    return -errno;
+    return UV__ERR(errno);
   /* Fall through. */
 #elif defined(__linux__)
   static int no_dup3;
@@ -1017,7 +1024,7 @@ int uv__dup2_cloexec(int oldfd, int newfd) {
     if (r != -1)
       return r;
     if (errno != ENOSYS)
-      return -errno;
+      return UV__ERR(errno);
     /* Fall through. */
     no_dup3 = 1;
   }
@@ -1033,7 +1040,7 @@ int uv__dup2_cloexec(int oldfd, int newfd) {
 #endif
 
     if (r == -1)
-      return -errno;
+      return UV__ERR(errno);
 
     err = uv__cloexec(newfd, 1);
     if (err) {
@@ -1048,29 +1055,16 @@ int uv__dup2_cloexec(int oldfd, int newfd) {
 
 int uv_os_homedir(char* buffer, size_t* size) {
   uv_passwd_t pwd;
-  char* buf;
   size_t len;
   int r;
 
-  if (buffer == NULL || size == NULL || *size == 0)
-    return -EINVAL;
+  /* Check if the HOME environment variable is set first. The task of
+     performing input validation on buffer and size is taken care of by
+     uv_os_getenv(). */
+  r = uv_os_getenv("HOME", buffer, size);
 
-  /* Check if the HOME environment variable is set first */
-  buf = getenv("HOME");
-
-  if (buf != NULL) {
-    len = strlen(buf);
-
-    if (len >= *size) {
-      *size = len + 1;
-      return -ENOBUFS;
-    }
-
-    memcpy(buffer, buf, len + 1);
-    *size = len;
-
-    return 0;
-  }
+  if (r != UV_ENOENT)
+    return r;
 
   /* HOME is not set, so call uv__getpwuid_r() */
   r = uv__getpwuid_r(&pwd);
@@ -1084,7 +1078,7 @@ int uv_os_homedir(char* buffer, size_t* size) {
   if (len >= *size) {
     *size = len + 1;
     uv_os_free_passwd(&pwd);
-    return -ENOBUFS;
+    return UV_ENOBUFS;
   }
 
   memcpy(buffer, pwd.homedir, len + 1);
@@ -1100,7 +1094,7 @@ int uv_os_tmpdir(char* buffer, size_t* size) {
   size_t len;
 
   if (buffer == NULL || size == NULL || *size == 0)
-    return -EINVAL;
+    return UV_EINVAL;
 
 #define CHECK_ENV_VAR(name)                                                   \
   do {                                                                        \
@@ -1130,7 +1124,7 @@ return_buffer:
 
   if (len >= *size) {
     *size = len + 1;
-    return -ENOBUFS;
+    return UV_ENOBUFS;
   }
 
   /* The returned directory should not have a trailing slash. */
@@ -1162,11 +1156,11 @@ int uv__getpwuid_r(uv_passwd_t* pwd) {
 
   getpwuid_r = dlsym(RTLD_DEFAULT, "getpwuid_r");
   if (getpwuid_r == NULL)
-    return -ENOSYS;
+    return UV_ENOSYS;
 #endif
 
   if (pwd == NULL)
-    return -EINVAL;
+    return UV_EINVAL;
 
   initsize = sysconf(_SC_GETPW_R_SIZE_MAX);
 
@@ -1183,7 +1177,7 @@ int uv__getpwuid_r(uv_passwd_t* pwd) {
     buf = uv__malloc(bufsize);
 
     if (buf == NULL)
-      return -ENOMEM;
+      return UV_ENOMEM;
 
     r = getpwuid_r(uid, &pw, buf, bufsize, &result);
 
@@ -1200,7 +1194,7 @@ int uv__getpwuid_r(uv_passwd_t* pwd) {
 
   if (result == NULL) {
     uv__free(buf);
-    return -ENOENT;
+    return UV_ENOENT;
   }
 
   /* Allocate memory for the username, shell, and home directory */
@@ -1211,7 +1205,7 @@ int uv__getpwuid_r(uv_passwd_t* pwd) {
 
   if (pwd->username == NULL) {
     uv__free(buf);
-    return -ENOMEM;
+    return UV_ENOMEM;
   }
 
   /* Copy the username */
@@ -1267,18 +1261,18 @@ int uv_os_getenv(const char* name, char* buffer, size_t* size) {
   size_t len;
 
   if (name == NULL || buffer == NULL || size == NULL || *size == 0)
-    return -EINVAL;
+    return UV_EINVAL;
 
   var = getenv(name);
 
   if (var == NULL)
-    return -ENOENT;
+    return UV_ENOENT;
 
   len = strlen(var);
 
   if (len >= *size) {
     *size = len + 1;
-    return -ENOBUFS;
+    return UV_ENOBUFS;
   }
 
   memcpy(buffer, var, len + 1);
@@ -1290,10 +1284,10 @@ int uv_os_getenv(const char* name, char* buffer, size_t* size) {
 
 int uv_os_setenv(const char* name, const char* value) {
   if (name == NULL || value == NULL)
-    return -EINVAL;
+    return UV_EINVAL;
 
   if (setenv(name, value, 1) != 0)
-    return -errno;
+    return UV__ERR(errno);
 
   return 0;
 }
@@ -1301,10 +1295,10 @@ int uv_os_setenv(const char* name, const char* value) {
 
 int uv_os_unsetenv(const char* name) {
   if (name == NULL)
-    return -EINVAL;
+    return UV_EINVAL;
 
   if (unsetenv(name) != 0)
-    return -errno;
+    return UV__ERR(errno);
 
   return 0;
 }
@@ -1321,17 +1315,17 @@ int uv_os_gethostname(char* buffer, size_t* size) {
   size_t len;
 
   if (buffer == NULL || size == NULL || *size == 0)
-    return -EINVAL;
+    return UV_EINVAL;
 
   if (gethostname(buf, sizeof(buf)) != 0)
-    return -errno;
+    return UV__ERR(errno);
 
   buf[sizeof(buf) - 1] = '\0'; /* Null terminate, just to be safe. */
   len = strlen(buf);
 
   if (len >= *size) {
     *size = len + 1;
-    return -ENOBUFS;
+    return UV_ENOBUFS;
   }
 
   memcpy(buffer, buf, len + 1);
@@ -1344,6 +1338,9 @@ uv_os_fd_t uv_get_osfhandle(int fd) {
   return fd;
 }
 
+int uv_open_osfhandle(uv_os_fd_t os_fd) {
+  return os_fd;
+}
 
 uv_pid_t uv_os_getpid(void) {
   return getpid();
@@ -1353,3 +1350,31 @@ uv_pid_t uv_os_getpid(void) {
 uv_pid_t uv_os_getppid(void) {
   return getppid();
 }
+
+
+int uv_os_getpriority(uv_pid_t pid, int* priority) {
+  int r;
+
+  if (priority == NULL)
+    return UV_EINVAL;
+
+  errno = 0;
+  r = getpriority(PRIO_PROCESS, (int) pid);
+
+  if (r == -1 && errno != 0)
+    return UV__ERR(errno);
+
+  *priority = r;
+  return 0;
+}
+
+
+int uv_os_setpriority(uv_pid_t pid, int priority) {
+  if (priority < UV_PRIORITY_HIGHEST || priority > UV_PRIORITY_LOW)
+    return UV_EINVAL;
+
+  if (setpriority(PRIO_PROCESS, (int) pid, priority) != 0)
+    return UV__ERR(errno);
+
+  return 0;
+}
